# Author: Anthony Stahler
# Last updated: Dec 16, 2023
# License: Free for personal use, for commercial use please reach out to the author via the BestEdit contact page: http://bestedit.net/contact/
# The most recent version of this script can be found at http://bestedit.net/premiere-pro-xml-fix/
#
# Version 2: added the ability to shift each file by a different number of frames, if desired
# Version 3: Correct instruction text
# Version 4: Fixed a bunch of bugs with the fta_ht functionality - note this script can now generate negative start values which is incorrect but seems to serve as a spacer when loaded in premiere so no reason to truncate to zero.
# Version 5: Fixed missing conversion for gain keyframes used to simulate exponential fades, and made a couple cosmetic fixes
# Version 6: Fix a negative IN value bug - audio will pad negative IN values with blank, but video will just refuse to shift negative IN, resulting in a desync between audio and video
# Version 7: Added support for additional framerates to suppliment the new xml-fix script functionality
# Version 8: Fix a bug with timeremap segments not shifting by the expected amount
#
# PLEASE NOTE: For simplicity this script assumes your XML file has been processed by the BestEdit Premiere Pro XML Fix script first. If it is not, this adjustment script will not work correctly!
#
# This script does not normalize the file names or do any of the other fix functionality found in the main XML Fix Script.
# Depending on the length of the source file you may also want to re-trim the very last clip in your project if it has too much footage or dead space at the end compared to what the original project intended
#

function FramesWithSpeedModifier
{
    # this converts a frame count taking into account a speed modifier, if none is present it just returns the original frame count
    param([long]$framelong, [decimal]$speeddecimal)

    #Write-Warning "REMAP: '$framelong','$speeddecimal'"
    if ($speeddecimal -ne -1)
    {
        $framelong = [math]::Round($framelong * $speeddecimal)
    }

    return [long]$framelong
}

# --------------------------
# VARIABLES 
# --------------------------

# this path has subfolders e.g. M:\movie-edits-exports\framerate-shift\output that contain xml files to be processed, this script will create these and the subfolders if you run it and they don't exist
$basepath = "M:\movie-edits-exports\framerate-shift\"

# set this value to your desired amount: 24 will shift it one second forward, -24 will shift one second back.  This will adjust all the ticks appropriately as well.
# set to zero if you want to use the granular fta_ht table with BestEdit File ID values
[long]$frames_to_adjust = 0

# if you leave frames_to_adjust set to zero the script will use this hashtable instead, which allows you to adjust different files by a different amount of frames (for projects with multiple files that are calibrated differently)
# the File ID you need for each file is typically on the BestEdit page for the edit,
# but you can also easily get a list of index for each file yourself, by using the list-project-file-indexes powershell script provided at http://bestedit.net/premiere-pro-xml-fix/
# You can run any edit through this script and it will list all the File ID indexes found in the xml file.
# the format to use in fta_ht is: "IndexNumber" =  FramesToShift;

# commenting out the default example values so we can warn the user if they forgot to put in a frames value
$fta_ht = @{
#"141" = 3; 
#"142" = 0;
#"143" = -5
}

# --------------------------
# CONSTANTS (don't touch)
# --------------------------
$targetext = "*.xml" 

# constant for 24fps 
[long]$pproconst = 10584000000
$fixclipconst = 1
$fixtransconst = 2

[long]$ppro23976fps = 10594584000 
[long]$ppro24fps = 10584000000 
[long]$ppro25fps = 10160640000 
[long]$ppro2997fps = 8475667200
[long]$ppro48fps = 5292000000
[long]$ppro50fps = 5080320000
[long]$ppro5994fps = 4237833600
[long]$ppro60fps = 4233600000 

# let's shove them in an array for ease of use
$fpsarr = @($ppro23976fps, $ppro24fps, $ppro25fps, $ppro2997fps, $ppro48fps, $ppro50fps, $ppro5994fps, $ppro60fps)
# IMPORTANT: keep this array the same length and order as $fpsarr!
$timebasearr = @(24,24,25,30,48,50,60,60)
$ntscarr = @('TRUE','FALSE','FALSE','TRUE','FALSE','FALSE','TRUE','FALSE')

# initializing this as 24fps, but it's going to change for each file processed
[long]$pproconst = $fpsarr[1]
$pprotimebase = $timebasearr[1]
$pprontsc = $ntscarr[1]

# --------------------------
# --------------------------


$srcpath = $basepath 
$destpath = $srcpath + "output\"
$finpath = $srcpath + "done\"

$silenceoutput = New-Item -ItemType Directory -Force -Path $destpath
$silenceoutput = New-Item -ItemType Directory -Force -Path $finpath




Get-ChildItem -path ( $srcpath + $targetext ) | foreach {
    $origname1 = $_.Name
    $origname1NOext = $_.BaseName
    $origext = $_.Extension

    $newname = $destpath + $origname1
    $donotmove = $true
    $negativein = $false

    if ($frames_to_adjust -ne 0 -or $fta_ht.Count -gt 0)
    {
        Write-Output "Processing '$origname1NOext'!"

        if ($frames_to_adjust -ne 0 -and $fta_ht.Count -gt 0)
        {
            Write-Output "WARNING: Shifting using universal frames_to_adjust value. If you meant to use granular fta_ht values please set frames_to_adjust = 0"
        }


        [decimal]$fileframerate = -1
        [long]$fileticksperframe = -1

        $trackactive = $false
        $transitionactive = $false
        [long]$tmpin = -1
        [long]$tmpout = -1
        [long]$tmpticksin = -1
        [long]$tmpticksout = -1
        $tmpaligned = $false
        $tmpstr = ""
        $currentfixtype = -1
        $isproject = $false

        # instead of the previous methods of processing each line we're going to grab the entire file so we can pre-process it in certain ways before doing the conversion pass
        $wholefile = [System.IO.File]::ReadAllText($srcpath + $origname1)

        # standardize newlines
        $wholefile = $wholefile -replace "`r`n", "`n"
        $wholefile = $wholefile -replace "`r", "`n"

        # The speed effect is hard to fix in the original format, we're going to regex it to make some changes and insert our own tags that make it easier to reprocess accurately
        # there's also a tag that indicates whether it's reverse or not, but we don't need that for our purposes here
        while ($wholefile -match '(?ms)(<clipitem ((?!<clipitem )(?!<besteditspeed>).)*?<effectid>timeremap</effectid>((?!<clipitem )(?!<besteditspeed>).)*?</clipitem>)')
        {
            # replace is a bad and messy way to do this, switching to Matches instead
            #$speedclip = $wholefile -replace '(?ms)(.*)(<clipitem ((?!<clipitem ).)*?(?!<besteditspeed>).*?<effectid>timeremap</effectid>.*?</clipitem>)(.*)', "`$2"
            $speedclip = $Matches.0
            #Write-Output "MATCH TIMEREMAP: '$speedclip'"

            $speedclip_orig = $speedclip

            # grab the multiplier for the speed
            #
            #<parameter authoringApp="PremierePro">
			#	<parameterid>speed</parameterid>
			#	<name>speed</name>
			#	<valuemin>-100000</valuemin>
			#	<valuemax>100000</valuemax>
			#	<value>92.3592</value>
			#</parameter>
            $speedparam = $speedclip -replace '(?ms)(.*)(<parameter ((?!<parameter ).)*?<parameterid>speed</parameterid>.*?<value>(.*?)</value>.*?</parameter>)(.*)', "`$4"
                    
            # Now attach the speed as a new tag right after the start of the clip, so we have it when we're processing the other variables
            # <clipitem id="clipitem-329">
            $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditspeed>$speedparam</besteditspeed>"

            # put our reprocessed speed clip back into the place of the original
            #$wholefile = $wholefile -replace $speedclip_orig, $speedclip
            $wholefile = $wholefile.Replace($speedclip_orig, $speedclip)
        }        

        $wholefile_lines = $wholefile -split "`n"

        # we're going to pre-read the first part of the file to grab the data to calculate the framerate
        #[System.IO.File]::ReadLines($srcpath + $origname1) | ForEach-Object {
        ForEach ($linedata_orig in $wholefile_lines) {

            if ($fileframerate -eq -1) # can't break out of a foreach-object, just read the rest of the lines quickly
            {
                #$linedata_orig = $_

                switch -regex ($linedata_orig)
                {
                    '(<clipitem )' { 
                        # we can use clips without transitions attached to easily calculate the exact framerate
                        $trackactive = $true
                        $transitionactive = $false
                        $tmpin = -1
                        $tmpout = -1
                        $tmpticksin = -1
                        $tmpticksout = -1
                        $tmpaligned = $true # set clips aligned automatically
                        break
                    }

                    '(</clipitem>)' { 
                        # Reset everything and wait for next clip if we didn't calculate the data we needed from this one
                        $trackactive = $false
                        $transitionactive = $false
                        $tmpin = -1
                        $tmpout = -1
                        $tmpticksin = -1
                        $tmpticksout = -1
                        $tmpaligned = $false
                        break
                    }

                    '(<transitionitem>)' { 
                        # we can also try to use audio transitions to calculate the framerate but the process is slightly different
                        $trackactive = $false
                        $transitionactive = $true
                        $tmpin = -1
                        $tmpout = -1
                        $tmpticksin = -1
                        $tmpticksout = -1
                        $tmpaligned = $false
                        break
                    }

                    '(</transitionitem>)' { 
                        # Reset everything and wait for next clip if we didn't calculate the data we needed from this one
                        $trackactive = $false
                        $transitionactive = $false
                        $tmpin = -1
                        $tmpout = -1
                        $tmpticksin = -1
                        $tmpticksout = -1
                        $tmpaligned = $false
                        break
                    }
                    

                    '(<in>)(.*)(</in>)' { 
                        if ($trackactive -eq $true)
                        {
                            $tmpstr = $linedata_orig -replace '(<in>)(.*)(</in>)', "`$2"
                            $tmpin = [System.Convert]::ToInt64($tmpstr)
                        }
                        break
                    }

                    '(<out>)(.*)(</out>)' { 
                        if ($trackactive -eq $true)
                        {
                            $tmpstr = $linedata_orig -replace '(<out>)(.*)(</out>)', "`$2"
                            $tmpout = [System.Convert]::ToInt64($tmpstr)
                        }    
                        break
                    }

                    '(<pproTicksIn>)(.*)(</pproTicksIn>)' { 
                        if ($trackactive -eq $true -or $transitionactive -eq $true)
                        {
                            $tmpstr = $linedata_orig -replace '(<pproTicksIn>)(.*)(</pproTicksIn>)', "`$2"
                            $tmpticksin = [System.Convert]::ToInt64($tmpstr)
                        }   
                        break
                    }

                    '(<pproTicksOut>)(.*)(</pproTicksOut>)' { 
                        if ($trackactive -eq $true -or $transitionactive -eq $true)
                        {
                            $tmpstr = $linedata_orig -replace '(<pproTicksOut>)(.*)(</pproTicksOut>)', "`$2"
                            $tmpticksout = [System.Convert]::ToInt64($tmpstr)
                        }   
                        break
                    }

                    '(<start>)(.*)(</start>)' { 
                        if ($trackactive -eq $true)
                        {
                            $tmpstr = $linedata_orig -replace '(<start>)(.*)(</start>)', "`$2"
                            if ([System.Convert]::ToInt64($tmpstr) -eq -1)
                            {
                                # There's some sort of padding involved in clips that have transitions attached, if this has a transition on either side we're ignoring it
                                $trackactive = $false
                                $transitionactive = $false
                                $tmpin = -1
                                $tmpout = -1
                                $tmpticksin = -1
                                $tmpticksout = -1
                                $tmpaligned = $false
                            }
                        }   

                        if ($transitionactive -eq $true)
                        {
                            # for transitions START is our IN tag
                            $tmpstr = $linedata_orig -replace '(<start>)(.*)(</start>)', "`$2"
                            $tmpin = [System.Convert]::ToInt64($tmpstr)
                        }
                        break
                    }

                    '(<end>)(.*)(</end>)' { 
                        if ($trackactive -eq $true)
                        {
                            $tmpstr = $linedata_orig -replace '(<end>)(.*)(</end>)', "`$2"
                            if ([System.Convert]::ToInt64($tmpstr) -eq -1)
                            {
                                # There's some sort of padding involved in clips that have transitions attached, if this has a transition on either side we're ignoring it
                                $trackactive = $false
                                $transitionactive = $false
                                $tmpin = -1
                                $tmpout = -1
                                $tmpticksin = -1
                                $tmpticksout = -1
                                $tmpaligned = $false
                            }
                        }   

                        if ($transitionactive -eq $true)
                        {
                            # for transitions END is our OUT tag
                            $tmpstr = $linedata_orig -replace '(<end>)(.*)(</end>)', "`$2"
                            $tmpout = [System.Convert]::ToInt64($tmpstr)
                        }
                        break
                    }

                    '(<alignment>)(.*)(</alignment>)' { 

                        if ($transitionactive -eq $true)
                        {
                            $tmpalign = $linedata_orig -replace '(<alignment>)(.*)(</alignment>)', "`$2"
                            $tmpalign = $tmpalign.ToLower().Trim()

                            if ($tmpalign -eq "center")
                            {
                                $tmpaligned = $true # center is the only alignment type that seems reliable, the rest are either slightly or majorly off
                            }
                                
                        }
                        break

                    }

                    '(<project>)' { 
                        # we only want sequences, not projects, warn the user
                        $isproject = $true
                        break
                    }

                }

                if ($tmpin -ne -1 -and $tmpout -ne -1 -and $tmpticksin -ne -1 -and $tmpticksout -ne -1 -and $tmpaligned -eq $true)
                {
                    if ($trackactive -eq $true)
                    {
                        # got all our variables, calculate it
                        $tmpout -= $tmpin
                        $tmpticksout -= $tmpticksin

                        $tmpticksout = $tmpticksout/$tmpout

                        $fileticksperframe = $tmpticksout
                        $fileframerate = ($ppro24fps/$tmpticksout)*24

                            
                    }
                    else
                    {
                        # using transitions we need a bit of fuzzy logic because there's a rounding error where the start tag sometimes needs to have 1 subtracted to be the correct frame count
                        # do it both ways and compare the result, if one is a decimal value then we know the other is the correct one.  If both are whole numbers toss them out and start again.
                        $tmpout -= $tmpin
                        $tmpticksout -= $tmpticksin

                        [decimal]$test1 = $tmpticksout/$tmpout

                        $tmpout++

                        [decimal]$test2 = $tmpticksout/$tmpout

                        if ((($test1 % 1) -eq 0) -and (($test2 % 1) -ne 0))
                        {
                            # test1 is correct value
                            $fileticksperframe = $test1
                            $fileframerate = ($ppro24fps/$test1)*24
                        } 
                        elseif ((($test2 % 1) -eq 0) -and (($test1 % 1) -ne 0))
                        {
                            # test2 is correct value
                            $fileticksperframe = $test2
                            $fileframerate = ($ppro24fps/$test2)*24
                        }
                        else
                        {
                            $fileticksperframe = -1
                            $fileframerate = -1
                        }

                        # if neither of those tests worked we either have two decimals or two whole numbers, toss them out and calculate the next one

                    }
                        
                    if ($fileframerate -ne -1)
                    {
                        # seems the clip division is rounded to 13 digits, round the transition math to the same length to keep everything standardized
                        $fileframerate = [math]::Round($fileframerate,13)
                    }

                    $trackactive = $false
                    $tmpin = -1
                    $tmpout = -1
                    $tmpticksin = -1
                    $tmpticksout = -1
                    $tmpaligned = $false
                }
            }
        }

        $pproconst = -1

        # we're just checking for exact matches here, unlike the XML fix script
        for ($i=0;$i -lt $fpsarr.Length;$i++)
        {
            if ($fileticksperframe -eq $fpsarr[$i])
            {
                $pproconst = $fpsarr[$i]
                $pprotimebase = $timebasearr[$i]
                $pprontsc = $ntscarr[$i]
                break
            }
        }

        if ($pproconst -ne -1)
        {
            # have to move these precalculate inside the loop now with the variable framerates
            # doesn't really save time any more, but leaving this way to avoid introducing bugs in the framerate changes

            # precalculate the ticks to save time
            $ticks_to_adjust = $frames_to_adjust * $pproconst

            $fta_ticks_ht = @{}

            # precalculate the hashtable ticks to save time
            foreach ($tmppair in $fta_ht.GetEnumerator())
            {
                [long]$tmpcalc = $tmppair.Value * $pproconst

                $fta_ticks_ht.Add($tmppair.Name, $tmpcalc)
            }

            $outstream = New-Object IO.StreamWriter $newname
        
            $currentfixtype = -1
            [long]$masterclip_id = -1
            [long]$linecount = -1
            [decimal]$clipspeed = -1

            $ingainblock = $false
            $videotrack = $true
            $outcache = ""

            ForEach ($linedata_orig in $wholefile_lines) 
            {
                #$linedata_orig = $_
                $linecount++

                [long]$tmplng = 0
                [long]$tmplng2 = 0

                switch -regex ($linedata_orig)
                {
                    '(<clipitem )' { 
                        $currentfixtype = $fixclipconst
                        $masterclip_id = -1
                        $clipspeed = -1
                    }

                    '(</clipitem>)' { 
                        $currentfixtype = -1
                        $masterclip_id = -1
                        $ingainblock = $false
                    }

                    '(<transitionitem>)' { 
                        $currentfixtype = $fixtransconst
                    }

                    '(</transitionitem>)' { 
                        $currentfixtype = -1
                        $ingainblock = $false
                    }

                    '(<masterclipid>)' { 
                        $tmpstr = $linedata_orig -replace '(<masterclipid>)(.*)(</masterclipid>)', "`$2"
                        $tmpstr = $tmpstr.ToLower().Trim()

                        # make the result just the number
                        $masterclip_id = $tmpstr.Replace("masterclip-", "")
                    }

                    '(<duration>)(.*)(</duration>)' { 
                        # we want to alter the duration slightly even if it isn't correct for the source, so that no Premiere bugs cause some of the edits to get cut off by duration values
                        # not sure if it ever matters, but better safe than sorry


                        if ($linecount -gt 10) # there is one duration tag at the start of the file that doesn't need to be adjusted
                        {

                            # only bother with it if the new source is longer than the original source
                            $tmplng = ($linedata_orig -replace '(<duration>)(.*)(</duration>)', "`$2").Trim()

                            if ($frames_to_adjust -ne 0)
                            {
                                if ($frames_to_adjust -gt 0)
                                {
                                    $tmplng += $frames_to_adjust # duration is the only one with +=, the rest are -= to make the user's shift directions more logical with negative numbers shifting frames left, positive right
                                }
                            }
                            else
                            {
                                # we're using the hashtable not the universal adjustment value
                                if ($masterclip_id -ne -1)
                                {
                                    if ($fta_ht.ContainsKey("$masterclip_id") -eq $true) 
                                    {
                                        $tmplng2 = $fta_ht["$masterclip_id"]

                                        if ($tmplng2 -gt 0) # once again, not trying to be precise, just making sure to only adjust up since this is just making sure the clips are as long or longer than we need, not too short
                                        {
                                            $tmplng += $tmplng2
                                        }
                                    }
                                    else
                                    {
                                        Write-Output "Error1: fta_ht did not contain a key matching '$masterclip_id'!"
                                    }
                                }
                                else
                                {

                                    Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                            
                                }
                            }

                            $linedata_orig = $linedata_orig -replace '(?<num1><duration>)(.*)(</duration>)', "`${num1}$tmplng`$2"
                        }
                    }

                    '(<start>)(.*)(</start>)' { 
                        # don't need to adjust start
                    }

                    '(<end>)(.*)(</end>)' { 
                        # don't need to adjust end
                    }

                    '(<cutPointTicks>)(.*)(</cutPointTicks>)' { 
                        # technically this is ticks but since we aren't adjusting framerate we shouldn't need to touch it, since it's an offset value, not an absolute value
                    }

                    '(<in>)(.*)(</in>)' { 
                        if ($currentfixtype -eq $fixclipconst)
                        {
                            $tmplng = ($linedata_orig -replace '(<in>)(.*)(</in>)', "`$2").Trim()
                        
                            if ($frames_to_adjust -ne 0)
                            {
                                $tmplng -= (FramesWithSpeedModifier $frames_to_adjust $clipspeed)
                            }
                            else
                            {
                                # we're using the hashtable not the universal adjustment value
                                if ($masterclip_id -ne -1)
                                {
                                    if ($fta_ht.ContainsKey("$masterclip_id") -eq $true)
                                    {
                                        $tmplng -= (FramesWithSpeedModifier $fta_ht["$masterclip_id"] $clipspeed)
                                    }
                                    else
                                    {
                                        Write-Output "Error2: fta_ht did not contain a key matching '$masterclip_id'!"
                                    }
                                }
                                else
                                {
                                    Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                                }
                            }

                            if ($tmplng -lt 0 -and $videotrack -eq $true)
                            {
                                # we only care about this if it's a video track, the audio tracks just fill with blank if they're negative
                                $negativein = $true
                                Write-Output "Error: negative IN value detected $linecount, correcting!"

                                $outcache_lines = $outcache -split "`n"

                                $startpos = -1
                                $nuketransition = $false
                                $intransiton = $false
                                $tmpend = -1
                                $tmpcpt = -1

                                # loop through the outcache to find the previous START and adjust it
                                for ($i=$outcache_lines.Length-1;$i -gt -1;$i--)
                                {
                                    switch -regex ($outcache_lines[$i])
                                    {
                                        '(<start>)(.*)(</start>)' { 
                                            # fix the start then exit this loop
                                            $tmpstart = ($outcache_lines[$i] -replace '(<start>)(.*)(</start>)', "`$2").Trim()
                                            # strange, the if ($tmpstart -gt -1) fails if $tmpstart is 0 unless we explicitly convert it to a number. Powershell is janky AF about conversions and comparisons
                                            $tmpstart = [System.Convert]::ToInt64($tmpstart)

                                            if ($intransiton -eq $false)
                                            {
                                                if ($tmpstart -gt -1)
                                                {
                                                    # subtract the negative value to add it to correct START
                                                    $tmpstart -= $tmplng
                                                    $outcache_lines[$i] = $outcache_lines[$i] -replace '(?<num1><start>)(.*)(</start>)', "`${num1}$tmpstart`$2"

                                                    $tmplng -= $tmplng
                                                    break
                                                }
                                                else
                                                {
                                                    # connected to a transition, look for that instead and break it
                                                    # this is going to be a little complicated since we have to parse the transition and figure out where the START would be without it, then adjust the START to that value plus the shift
                                                    $startpos = $i
                                                    $nuketransition = $true
                                                }
                                            }
                                            else
                                            {
                                                # start value of the previous transition, should only hit this if the clip was connected to a transition
                                                if ($tmpend -ne -1 -and $tmpcpt -ne -1 -and $startpos -ne -1)
                                                {
                                                    # add any shift from the cutpointticks to the start of the transition so that we have the correct un-shift start value
                                                    $tmpstart += [math]::Round($tmpcpt/$pproconst)

                                                    # now shift the start
                                                    $tmpstart -= $tmplng
                                                    $tmplng -= $tmplng

                                                    $outcache_lines[$startpos] = $outcache_lines[$startpos] -replace '(?<num1><start>)(.*)(</start>)', "`${num1}$tmpstart`$2"
                                                }
                                                else
                                                {
                                                    Write-Output "Error: could not parse transition connected to negative IN. $linecount!"
                                                
                                                }

                                                break
                                            }
                                        }

                                        '(<end>)(.*)(</end>)' { 
                                            if ($intransiton -eq $true)
                                            {
                                                $tmpend = ($outcache_lines[$i] -replace '(<end>)(.*)(</end>)', "`$2").Trim()
                                            }
                                        }

                                        '(<cutPointTicks>)(.*)(</cutPointTicks>)' { 
                                            if ($intransiton -eq $true)
                                            {
                                                $tmpcpt = ($outcache_lines[$i] -replace '(<cutPointTicks>)(.*)(</cutPointTicks>)', "`$2").Trim()
                                            }
                                        }
                                    
                                        '(<transitionitem>)' { 
                                            $intransiton = $false

                                            # if we're already nuking then finish it up and get out of this loop
                                            if ($nuketransition -eq $true)
                                            {
                                                $outcache_lines[$i] = ""
                                                $nuketransition = $false
                                                break
                                            }
                                        }

                                        '(</transitionitem>)' { 
                                            $intransiton = $true
                                        }
                                    }

                                    # inside the previous transition, nuke this line to erase it
                                    if ($nuketransition -eq $true -and $intransiton -eq $true)
                                    {
                                        $outcache_lines[$i] = ""
                                    }
                                }

                                $outcache = $outcache_lines -join "`n"
                            }

                        

                            $linedata_orig = $linedata_orig -replace '(?<num1><in>)(.*)(</in>)', "`${num1}$tmplng`$2"
                        }

                        $donotmove = $false
                    }

                    '(<out>)(.*)(</out>)' { 
                        if ($currentfixtype -eq $fixclipconst)
                        {
                            $tmplng = ($linedata_orig -replace '(<out>)(.*)(</out>)', "`$2").Trim()
                        
                            if ($frames_to_adjust -ne 0)
                            {
                                $tmplng -= (FramesWithSpeedModifier $frames_to_adjust $clipspeed)
                            }
                            else
                            {
                                # we're using the hashtable not the universal adjustment value
                                if ($masterclip_id -ne -1)
                                {
                                    if ($fta_ht.ContainsKey("$masterclip_id") -eq $true)
                                    {
                                        $tmplng -= (FramesWithSpeedModifier $fta_ht["$masterclip_id"] $clipspeed)
                                    }
                                    else
                                    {
                                        Write-Output "Error3: fta_ht did not contain a key matching '$masterclip_id'!"
                                    }
                                }
                                else
                                {
                                    Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                                }
                            }

                            $linedata_orig = $linedata_orig -replace '(?<num1><out>)(.*)(</out>)', "`${num1}$tmplng`$2"
                        }
                    }

                    '(<pproTicksIn>)(.*)(</pproTicksIn>)' { 
                        if ($currentfixtype -eq $fixclipconst)
                        {
                            $tmplng = ($linedata_orig -replace '(<pproTicksIn>)(.*)(</pproTicksIn>)', "`$2").Trim()
                        
                            if ($frames_to_adjust -ne 0)
                            {
                                $tmplng -= $ticks_to_adjust
                            }
                            else
                            {
                                # we're using the hashtable not the universal adjustment value
                                if ($masterclip_id -ne -1)
                                {
                                    if ($fta_ticks_ht.ContainsKey("$masterclip_id") -eq $true)
                                    {
                                        $tmplng -= $fta_ticks_ht["$masterclip_id"]
                                    }
                                    else
                                    {
                                        Write-Output "Error4: fta_ticks_ht did not contain a key matching '$masterclip_id'!"
                                    }
                                }
                                else
                                {
                                    Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                                }
                            }

                            if ($tmplng -lt 0 -and $videotrack -eq $true)
                            {
                                $tmplng = 0
                            }

                            $linedata_orig = $linedata_orig -replace '(?<num1><pproTicksIn>)(.*)(</pproTicksIn>)', "`${num1}$tmplng`$2"
                        }
                    }

                    '(<pproTicksOut>)(.*)(</pproTicksOut>)' { 
                        if ($currentfixtype -eq $fixclipconst)
                        {
                            $tmplng = ($linedata_orig -replace '(<pproTicksOut>)(.*)(</pproTicksOut>)', "`$2").Trim()
                        
                            if ($frames_to_adjust -ne 0)
                            {
                                $tmplng -= $ticks_to_adjust
                            }
                            else
                            {
                                # we're using the hashtable not the universal adjustment value
                                if ($masterclip_id -ne -1)
                                {
                                    if ($fta_ticks_ht.ContainsKey("$masterclip_id") -eq $true)
                                    {
                                        $tmplng -= $fta_ticks_ht["$masterclip_id"]
                                    }
                                    else
                                    {
                                        Write-Output "Error5: fta_ht did not contain a key matching '$masterclip_id'!"
                                    }
                                }
                                else
                                {
                                    Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                                }
                            }

                            $linedata_orig = $linedata_orig -replace '(?<num1><pproTicksOut>)(.*)(</pproTicksOut>)', "`${num1}$tmplng`$2"   
                        }    

                        # done with the clipspeed after altering in/out, clear it out so we don't accidentally impact any filters or anything that might share these tags
                        $clipspeed = -1
                    }

                    '(<effecttype>)(.*)(</effecttype>)' { 
                        $tmpstr = ($linedata_orig -replace '(<effecttype>)(.*)(</effecttype>)', "`$2").Trim()

                        if ($tmpstr -eq "audiolevels")
                        {
                            #<effecttype>audiolevels</effecttype>
                            $ingainblock = $true
                        }
                    }

                    '(</effect>)' { 
                        $ingainblock = $false
                    }

                    '(<when>)(.*)(</when>)' { 

                        if ($ingainblock -eq $true)
                        {
                            $tmplng = ($linedata_orig -replace '(<when>)(.*)(</when>)', "`$2").Trim()

                            # same code as shifting in/out block
                            if ($frames_to_adjust -ne 0)
                            {
                                $tmplng -= $frames_to_adjust
                            }
                            else
                            {
                                # we're using the hashtable not the universal adjustment value
                                if ($masterclip_id -ne -1)
                                {
                                    if ($fta_ht.ContainsKey("$masterclip_id") -eq $true)
                                    {
                                        $tmplng -= $fta_ht["$masterclip_id"]
                                    }
                                    else
                                    {
                                        Write-Output "Error3: fta_ht did not contain a key matching '$masterclip_id'!"
                                    }
                                }
                                else
                                {
                                    Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                                }
                            }

                            $linedata_orig = $linedata_orig -replace '(?<num1><when>)(.*)(</when>)', "`${num1}$tmplng`$2"
                        }
                    }

                    '(<track )(.*)(currentExplodedTrackIndex)(.*)(totalExplodedTrackCount)' { 
                        # audio track
                        $videotrack = $false
                    }

                    '(</track>)' { 
                        # default back to video track until we detect another audio track
                        $videotrack = $true
                    }

                    '(<besteditspeed>)(.*)(</besteditspeed>)' { 
                            
                        $clipspeed = [System.Convert]::ToDecimal(($linedata_orig -replace '(<besteditspeed>)(.*)(</besteditspeed>)', "`$2").Trim())
                        # convert to percent, the original tag has it as a whole number
                        $clipspeed = $clipspeed / 100

                         # not converting to a percent for the shift script, we need the whole number multiplier for the in/out modifier
                         # 10 / 100 = 0.1
                         # 1 / 0.1 = 10
                         #
                         # 50 / 100 = 0.5
                         # 1 / 0.5 = 2
                         $clipspeed = 1 / $clipspeed

                         # need to investigate what to do with this if it's a fractional amount like a 33% speed change
                         #<value>33.33</value> = slowed to a third speed
                         #<value>337.361</value> = sped up random triple speed
                         # looks like the in/out values have rounding applied, need to be careful
                         # 13737/3.37361 = 4071.899241465374 = <in>4072</in>
                         #
                         # I don't think there's any good way to compensate because we don't know the original IN value, so we just have to
                         # make peace with the fact that it may be a frame off due to multiplying, shifting, and dividing and rounding again if it isn't an even speed value.

                        $linedata_orig = '' # nuke this line so we don't insert our temp tag in the final project
                        break
                    }
                }


                # because of the bug in the video IN shifts we need to continually cache any non-audio data until we're sure it isn't a problem
                if ($videotrack -eq $false)
                {
                    if ($outcache -ne "")
                    {
                        $outstream.WriteLine($outcache)
                        $outcache = ""
                    }

                    $outstream.WriteLine($linedata_orig)
                }
                else
                {
                    $outcache += "$linedata_orig`n"

                    if ($negativein -eq $true)
                    {
                        # if we just fixed a negative IN, flush the cache just to be safe and avoid any potential bugs
                        $outstream.WriteLine($outcache)
                        $outcache = ""

                        $negativein = $false
                    }
                }
            }

            if ($outcache -ne "")
            {
                $outstream.WriteLine($outcache)
                $outcache = ""
            }

            $outstream.Close()
            
        
            if ($donotmove -eq $false)
            {
                # current time with a bit of randomness to avoid getting the same number over and over since we process fast
                $currenttime = [math]::Floor( (New-TimeSpan -Start (Get-Date "01/01/1970") -End (Get-Date)).TotalSeconds ) - (Get-Random) 

                # fix negative numbers
                if ($currenttime -lt 0)
                {
                    $currenttime = -$currenttime
                }

                # zip the created file and delete it
                Compress-Archive -Path $newname -DestinationPath ($destpath + $origname1NOext + "_" + $currenttime + ".zip")
                Remove-Item -LiteralPath ($newname) -Force -Recurse
            }
            else
            {
                Remove-Item -LiteralPath ($newname) -Force -Recurse
            }
        }
    }
    else
    {
        Write-Output "Can't adjust files by 0 frames, please set the `$frames_to_adjust variable to a positive or negative number to shift your project!"
    }

    if ($donotmove -eq $false)
    {
        Get-Item -LiteralPath ( $srcpath + $origname1 ) | Move-Item -Force -Destination { $finpath + $origname1 }
    }
    else
    {
        Write-Output "Found nothing to adjust in '$origname1'!"
    }
}
